/* -*- mode: C++; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

#include <iostream>
#include <fstream>
#include <sstream>
#include <assert.h>

#include "integer.cuh"
#include "gfp_gpu.cuh"
#include "gpu_memory.cuh"
#include "kernel_fp.cuh"
#include "kernel_ec.cuh"

#include "timing.h"

using namespace std;

#define CUDADEVICE 1

uint64_t t_start,t_end;
double tps;

// -------------------------------------------------------------------
#define BENCH(__N__,__CMD__)		       \
  t_start = microseconds();		       \
  for (unsigned long  i = 0; i < __N__; ++i) { \
    asm volatile ("#begin ");		       \
    __CMD__				       \
      asm volatile ("#end ");		       \
  }					       \
  t_end = microseconds();		       \
  tps = (double)(t_end - t_start);	       \
  tps /= (double)(__N__);                      


/************************************************************************************************/
/****************** BENCHMARK LAUNCH FUNCTIONS FOR POINT MULTIPLICATION TESTS *******************/
/************************************************************************************************/

#ifdef BENCHMARK_POINT_MULTIPLICATION

/************************************************************************************************/

void get_GPU_data(GPU_vector_data<N> gpu_data, 
		  vector<Integer<N> >& v_Qx,
		  vector<Integer<N> >& v_Qy,
		  vector<Integer<N> >& v_Qz
		  ) 
{
	//copy result from device to host
	Memory<N, NUMBER_THREADS>::sendToCPU(v_Qx, gpu_data.__v_Qx);
	Memory<N, NUMBER_THREADS>::sendToCPU(v_Qy, gpu_data.__v_Qy);
	Memory<N, NUMBER_THREADS>::sendToCPU(v_Qz, gpu_data.__v_Qz);

}

/************************************************************************************************/

//write a line in file_name (result file) for this execution, including N, number of fp, curve, point and k and the times
inline void write_result(char* file_name, 
			 string param,
			 float execTime,
			 double totalTime
			 ) 
{
	fstream file(file_name,ios::in | ios::out | ios::ate);
	if(!file) {
		ofstream file_cstr(file_name);
		file_cstr.close();
		file.open(file_name, ios::in | ios::out | ios::ate);
		file << "All time are in microseconds" << endl;
		file << "# bits:# Fp:# Curve:# Point:# k: HW(average):Time of Computation:For one [k]P:Total Time:For one [k]P"<<endl;
	}
	file << N << ":" << param << ":" << 1000*execTime << ":" << 1000*execTime/NUMBER_THREADS << ":" << totalTime << ":" << (totalTime/NUMBER_THREADS) << endl;
	file.close();
}

/************************************************************************************************/

//read benchmark file and send data to the GPU. Return a string representing the number of Fp:curve:point:k
inline string read_data(char* file_name,
			vector<Integer<N> >& v_prime,
			vector<Integer<N> >& v_a, 
			vector<uint>&        v_invp, 
			vector<Integer<N> >& v_Px, 
			vector<Integer<N> >& v_Py, 
			vector<Integer<N> >& v_Pz, 
			vector<Integer<N> >& v_k,
			vector<Integer<N> >& v_Qx,
			vector<Integer<N> >& v_Qy,
			vector<Integer<N> >& v_Qz
			)
{

	/*************************************
	 * READ THE DATA FROM BENCHMARK FILE *
	 *************************************/
	
	//number of different Fp
	int nb_fp;
	//number of different curves per Fp
	int nb_curve;
	//number of different points per curve
	int nb_point;
	//number of different k per point
	int nb_k;

	ifstream file(file_name);
	istringstream is;
	string line;

	//get number of Fp
	getline(file, line);
	is.str(line);
	is >> nb_fp;
	is.clear();

	//get number of curves
	getline(file, line);
	is.str(line);
	is >> nb_curve;
	is.clear();

	//get number of point per curve
	getline(file, line);
	is.str(line);
	is >> nb_point;
	is.clear();

	//get number of k
	getline(file, line);
	is.str(line);
	is >> nb_k;
	is.clear();

	assert(NUMBER_THREADS == nb_fp*nb_curve*nb_point*nb_k);

	//Number threads per Fp
	uint nbThreadsFp    = nb_curve*nb_point*nb_k;
	//Number threads per curve
	uint nbThreadsCurve = nb_point*nb_k;
	//Number threads per point
	uint nbThreadsPoint = nb_k;

	uint hamming_weight_sum = 0;
	//point coordinates are in affine form in file - changing to Jacobian form assuming z = 1
	Integer<N> z("1");

	for(uint i = 0; i < nb_fp; i++) {
		//Get the ith Fp (representing by its prime)
		getline(file, line);		
		GFp<N> fp(line);

		//Initialize (number threads per Fp) Integer<N> to the GFp value
		for(uint j = 0; j < nbThreadsFp; j++) {
		       	v_prime[i*nbThreadsFp+j] = fp._prime;
			v_invp[i*nbThreadsFp+j]  = fp._invp;
		}

		//Get curve(s) (parameter a) for the current Fp
		for(uint j = 0; j < nb_curve; j++) {
			getline(file, line);
			Integer<N> a(line);
			//Initialize (number threads per curve) Integer<N> to the a value
			for(uint k = 0; k <nbThreadsCurve; k++) {
				v_a[i*nbThreadsFp + j*nbThreadsCurve + k] = a;
			}
			//Get point(s) for the current curve
			for(uint k = 0; k < nb_point; k++) {
				getline(file, line);
				Integer<N> x(line);
				getline(file, line);
				Integer<N> y(line);
				for(uint l = 0; l < nbThreadsPoint; l++) {
					v_Px[i*nbThreadsFp + j*nbThreadsCurve + k*nbThreadsPoint + l] = x;
					v_Py[i*nbThreadsFp + j*nbThreadsCurve + k*nbThreadsPoint + l] = y;
					v_Pz[i*nbThreadsFp + j*nbThreadsCurve + k*nbThreadsPoint + l] = z;
				}
				//Get k for the current point
				for(uint l = 0; l < nb_k; l++) {
					getline(file, line);
					Integer<N> kin(line);
					hamming_weight_sum += kin.get_hamming_weight();
					v_k[i * nbThreadsFp + j*nbThreadsCurve + k*nbThreadsPoint + l] = kin;
				}
			}
		}
	}
	file.close();
	ostringstream os;
	os << nb_fp << ":" << nb_curve << ":" << nb_point << ":" << nb_k << ":" << hamming_weight_sum / NUMBER_THREADS;
	return os.str();
}

/************************************************************************************************/

inline void alloc_send_data (GPU_vector_data<N>&  gpu_data,
			     vector<Integer<N> >& v_prime,
			     vector<Integer<N> >& v_a, 
			     vector<uint>&        v_invp, 
			     vector<Integer<N> >& v_Px, 
			     vector<Integer<N> >& v_Py, 
			     vector<Integer<N> >& v_Pz, 
			     vector<Integer<N> >& v_k,
			     vector<Integer<N> >& v_Qx, 
			     vector<Integer<N> >& v_Qy, 
			     vector<Integer<N> >& v_Qz 
			     )		      
{
	/******************************
	 * ALLOCATION AND COPY TO GPU *
	 ******************************/

	// allocate space on the GPU for [k]P results
	Memory<N, NUMBER_THREADS>::allocGPU(gpu_data.__v_Qx);
	Memory<N, NUMBER_THREADS>::allocGPU(gpu_data.__v_Qy);
	Memory<N, NUMBER_THREADS>::allocGPU(gpu_data.__v_Qz);

	// send the data to the GPU
	Memory<N, NUMBER_THREADS>::sendToGPU(gpu_data.__v_k, v_k);
	Memory<N, NUMBER_THREADS>::sendToGPU(gpu_data.__v_a, v_a);
	Memory<N, NUMBER_THREADS>::sendToGPU(gpu_data.__v_prime, v_prime);
	Memory<N, NUMBER_THREADS>::sendToGPU(gpu_data.__v_invp, v_invp);
	Memory<N, NUMBER_THREADS>::sendToGPU(gpu_data.__v_Px, v_Px);
	Memory<N, NUMBER_THREADS>::sendToGPU(gpu_data.__v_Py, v_Py);
	Memory<N, NUMBER_THREADS>::sendToGPU(gpu_data.__v_Pz, v_Pz);
}

/************************************************************************************************/

#endif

/************************************************************************************************/
/********************* BENCHMARK LAUNCH FUNCTIONS FOR FP ARITHMETIC TESTS ***********************/
/************************************************************************************************/

#ifdef BENCHMARK_FP_ARITHMETIC

void get_GPU_data(GPU_vector_data<N> gpu_data, 
		  vector<Integer<N> >& v_Qx,
		  vector<Integer<N> >& v_Qy,
		  vector<Integer<N> >& v_Qz
		  ) 
{
	//copy result from device to host
	Memory<N, NUMBER_THREADS>::sendToCPU(v_Qx, gpu_data.__v_Px);
}

/************************************************************************************************/

inline void write_result(char* file_name,
			 string param,
			 float execTime, 
			 double totalTime
			 )
{
	fstream file(file_name,ios::in | ios::out | ios::ate);
	if(!file) {
		ofstream file_cstr(file_name);
		file_cstr.close();
		file.open(file_name, ios::in | ios::out | ios::ate);
		string op("");
		file << "All time are in microseconds" << endl;
		#ifdef FP_ADDITION
		op = "addition";
		#endif
		#ifdef FP_MULTIPLICATION
		op = "multiplication";
		#endif
		file << "# bits:# operations:Time of Computation (ms):For one " << op <<" :Total Time (ms):For one " << op <<endl;
	}
	file << N << ":" << NB_OPERATIONS << ":" << execTime*1000 << ":" << 1000*execTime/(NUMBER_THREADS*NB_OPERATIONS) << ":" << totalTime << ":" << (totalTime/(NUMBER_THREADS*NB_OPERATIONS)) << endl;
	file.close();
}

/************************************************************************************************/

inline string read_data(char* file_name,
			vector<Integer<N> >& v_prime,
			vector<Integer<N> >& v_a, 
			vector<uint>&        v_invp, 
			vector<Integer<N> >& v_Px, 
			vector<Integer<N> >& v_Py, 
			vector<Integer<N> >& v_Pz, 
			vector<Integer<N> >& v_k,
			vector<Integer<N> >& v_Qx,
			vector<Integer<N> >& v_Qy,
			vector<Integer<N> >& v_Qz
		      ) 
{
	/*************************************
	 * READ THE DATA FROM BENCHMARK FILE *
	 *************************************/
	
	//number of different Fp
	int nb_fp;

	ifstream file(file_name);
	istringstream is;
	string line;

	//get number of Fp
	getline(file, line);
	is.str(line);
	is >> nb_fp;
	is.clear();

	assert(NUMBER_THREADS == nb_fp);
	//Get the ith Fp (representing by its prime)
	getline(file, line);		
	GFp<N> fp(line);
	for(uint i = 0; i < nb_fp; i++) {
	       	v_prime[i] = fp._prime;
		v_invp[i]  = fp._invp;
		getline(file, line);
		Integer<N> op1(line);
		v_Px[i] = op1;
		getline(file, line);
		Integer<N> op2(line);
		v_Py[i] = op2;
	}
	file.close();
	return "";
}

/************************************************************************************************/

inline void alloc_send_data(GPU_vector_data<N>&  gpu_data,
			    vector<Integer<N> >& v_prime,
			    vector<Integer<N> >& v_a, 
			    vector<uint>&        v_invp, 
			    vector<Integer<N> >& v_Px, 
			    vector<Integer<N> >& v_Py, 
			    vector<Integer<N> >& v_Pz, 
			    vector<Integer<N> >& v_k,
			    vector<Integer<N> >& v_Qx, 
			    vector<Integer<N> >& v_Qy, 
			    vector<Integer<N> >& v_Qz
) 
{
	// allocate space on the GPU for [k]P results
	Memory<N, NUMBER_THREADS>::allocGPU(gpu_data.__v_Qx);

	// send the data to the GPU
	Memory<N, NUMBER_THREADS>::sendToGPU(gpu_data.__v_prime, v_prime);
	Memory<N, NUMBER_THREADS>::sendToGPU(gpu_data.__v_invp, v_invp);
	Memory<N, NUMBER_THREADS>::sendToGPU(gpu_data.__v_Px, v_Px);
	Memory<N, NUMBER_THREADS>::sendToGPU(gpu_data.__v_Py, v_Py);

}

#endif

/************************************************************************************************/
/****************** BENCHMARK LAUNCH FUNCTIONS FOR ELLIPTIC CURVE ARITHMETIC ********************/
/************************************************************************************************/

#ifdef BENCHMARK_EC_ARITHMETIC

void get_GPU_data(GPU_vector_data<N> gpu_data, 
		  vector<Integer<N> >& v_Qx,
		  vector<Integer<N> >& v_Qy,
		  vector<Integer<N> >& v_Qz
		  ) 
{
	//copy result from device to host
	Memory<N, NUMBER_THREADS>::sendToCPU(v_Qx, gpu_data.__v_Px);
	Memory<N, NUMBER_THREADS>::sendToCPU(v_Qy, gpu_data.__v_Py);
	Memory<N, NUMBER_THREADS>::sendToCPU(v_Qz, gpu_data.__v_Pz);

}

/************************************************************************************************/

//write a line in file_name (result file) for this execution, including N, number of fp, curve, point and k and the times
inline void write_result(char* file_name, string param, float execTime, double totalTime) {
	fstream file(file_name,ios::in | ios::out | ios::ate);
	if(!file) {
		ofstream file_cstr(file_name);
		file_cstr.close();
		file.open(file_name, ios::in | ios::out | ios::ate);
		file << "All time are in microseconds" << endl;
		#ifdef EC_ADDITION
		file << "# bits:# Fp:# Curves:# Points:# Operations:Time of Computation:For one addition:Total Time:For one addition"<<endl;
		#endif
		#ifdef EC_DOUBLING
		file << "# bits:# Fp:# Curves:# Points:# Operations:Time of Computation:For one doubling:Total Time:For one doubling"<<endl;
		#endif
	}
	file << N << ":" << param <<":" << NB_OPERATIONS << ":" << 1000*execTime << ":" << 1000*execTime/(NUMBER_THREADS*NB_OPERATIONS) << ":" << totalTime << ":" << (totalTime/(NUMBER_THREADS*NB_OPERATIONS)) << endl;
	file.close();
}

/************************************************************************************************/

//read benchmark file and send data to the GPU. Return a string representing the number of Fp:curve:point:k
inline string read_data(char* file_name,
			vector<Integer<N> >& v_prime,
			vector<Integer<N> >& v_a, 
			vector<uint>&        v_invp, 
			vector<Integer<N> >& v_Px, 
			vector<Integer<N> >& v_Py, 
			vector<Integer<N> >& v_Pz,
			vector<Integer<N> >& v_k,
			vector<Integer<N> >& v_Qx,
			vector<Integer<N> >& v_Qy,
			vector<Integer<N> >& v_Qz
			) 
{

	/*************************************
	 * READ THE DATA FROM BENCHMARK FILE *
	 *************************************/
	
	//number of different Fp
	int nb_fp;
	//number of different curves per Fp
	int nb_curve;
	//number of different points per curve
	int nb_point;

	ifstream file(file_name);
	istringstream is;
	string line;

	//get number of Fp
	getline(file, line);
	is.str(line);
	is >> nb_fp;
	is.clear();

	//get number of curves
	getline(file, line);
	is.str(line);
	is >> nb_curve;
	is.clear();

	//get number of point per curve
	getline(file, line);
	is.str(line);
	is >> nb_point;
	is.clear();

	assert(NUMBER_THREADS == nb_fp*nb_curve*nb_point);

	//Number threads per Fp
	uint nbThreadsFp    = nb_curve*nb_point;
	//Number threads per curve
	uint nbThreadsCurve = nb_point;

	//point coordinates are in affine form in file - changing to Jacobian form assuming z = 1
	Integer<N> z("1");

	for(uint i = 0; i < nb_fp; i++) {
		//Get the ith Fp (representing by its prime)
		getline(file, line);		
		GFp<N> fp(line);

		//Initialize (number threads per Fp) Integer<N> to the GFp value
		for(uint j = 0; j < nbThreadsFp; j++) {
		       	v_prime[i*nbThreadsFp+j] = fp._prime;
			v_invp[i*nbThreadsFp+j]  = fp._invp;
		}

		//Get curve(s) (parameter a) for the current Fp
		for(uint j = 0; j < nb_curve; j++) {
			getline(file, line);
			Integer<N> a(line);
			//Initialize (number threads per curve) Integer<N> to the a value
			for(uint k = 0; k <nbThreadsCurve; k++) {
				v_a[i*nbThreadsFp + j*nbThreadsCurve + k] = a;
			}
			//Get point(s) for the current curve
			for(uint k = 0; k < nb_point; k++) {
				getline(file, line);
				Integer<N> x(line);
				getline(file, line);
				Integer<N> y(line);
				
				v_Px[i*nbThreadsFp + j*nbThreadsCurve + k] = x;
				v_Py[i*nbThreadsFp + j*nbThreadsCurve + k] = y;
				v_Pz[i*nbThreadsFp + j*nbThreadsCurve + k] = z;
#ifdef EC_ADDITION
				getline(file, line);
				Integer<N> xQ(line);
				getline(file, line);
				Integer<N> yQ(line);
				v_Qx[i*nbThreadsFp + j*nbThreadsCurve + k] = xQ;
				v_Qy[i*nbThreadsFp + j*nbThreadsCurve + k] = yQ;
				v_Qz[i*nbThreadsFp + j*nbThreadsCurve + k] = z;
#endif
#ifdef EC_DOUBLING
				getline(file, line);
				getline(file, line);
#endif
				
			}
		}
	}
	file.close();
	ostringstream os;
	os << nb_fp << ":" << nb_curve << ":" <<nb_point;
	return os.str();
}

/************************************************************************************************/

inline void alloc_send_data (GPU_vector_data<N>&  gpu_data,
			     vector<Integer<N> >& v_prime,
			     vector<Integer<N> >& v_a, 
			     vector<uint>&        v_invp, 
			     vector<Integer<N> >& v_Px, 
			     vector<Integer<N> >& v_Py, 
			     vector<Integer<N> >& v_Pz, 
			     vector<Integer<N> >& v_k,
			     vector<Integer<N> >& v_Qx,
			     vector<Integer<N> >& v_Qy,
			     vector<Integer<N> >& v_Qz
			     )		      
{
	/******************************
	 * ALLOCATION AND COPY TO GPU *
	 ******************************/
	// send the data to the GPU
#ifdef EC_ADDITION
	Memory<N, NUMBER_THREADS>::sendToGPU(gpu_data.__v_Qx, v_Qx);
	Memory<N, NUMBER_THREADS>::sendToGPU(gpu_data.__v_Qy, v_Qy);
	Memory<N, NUMBER_THREADS>::sendToGPU(gpu_data.__v_Qz, v_Qz);
#endif
	Memory<N, NUMBER_THREADS>::sendToGPU(gpu_data.__v_a, v_a);
	Memory<N, NUMBER_THREADS>::sendToGPU(gpu_data.__v_prime, v_prime);
	Memory<N, NUMBER_THREADS>::sendToGPU(gpu_data.__v_invp, v_invp);
	Memory<N, NUMBER_THREADS>::sendToGPU(gpu_data.__v_Px, v_Px);
	Memory<N, NUMBER_THREADS>::sendToGPU(gpu_data.__v_Py, v_Py);
	Memory<N, NUMBER_THREADS>::sendToGPU(gpu_data.__v_Pz, v_Pz);
}

/************************************************************************************************/

#endif

/********************************************/
/* COMMON FUNCTIONS FOR LAUNCHING BENCHMARK */
/********************************************/

/*
  PROTOTYPES OF FUNCTIONS TO BE OVERLOADED :
  * string read_data(char*,vector<Integer<N> >&, vector<Integer<N> >&, vector<uint>&, vector<Integer<N> >&, vector<Integer<N> >&, vector<Integer<N> >&, vector<Integer<N> >&, vector<Integer<N> >&, vector<Integer<N> >&, vector<Integer<N> >&) : read data from char* file and fill vectors of data
  * void alloc_send_data(GPU_vector_data<N>&, vector<Integer<N> >&, vector<Integer<N> >&, vector<uint>&, vector<Integer<N> >&, vector<Integer<N> >&, vector<Integer<N> >&, vector<Integer<N> >&, vector<Integer<N> >&, vector<Integer<N> >&, vector<Integer<N> >&) : send data to the GPU
  * get_data(GPU_vector_data<N>&, vector<Integer<N> >&, vector<Integer<N> >&, vector<Integer<N> >&) : read data from the GPU
  * write_result(char*, string, float, double)

*/

//Allocate and copy data to the device, run kernel function and get the result to the host
inline float run_execution(vector<Integer<N> >& v_prime,
			   vector<Integer<N> >& v_a, 
			   vector<uint>&        v_invp, 
			   vector<Integer<N> >& v_Px, 
			   vector<Integer<N> >& v_Py, 
			   vector<Integer<N> >& v_Pz, 
			   vector<Integer<N> >& v_k,
			   vector<Integer<N> >& v_Qx,
			   vector<Integer<N> >& v_Qy,
			   vector<Integer<N> >& v_Qz
			   )
{
	//Choose cuda device
	cudaSetDevice(CUDADEVICE);

	// define the data on GPU 
	GPU_vector_data<N> gpu_data;

	alloc_send_data(gpu_data, v_prime, v_a, v_invp, v_Px, v_Py, v_Pz, v_k, v_Qx, v_Qy, v_Qz);

	//Dimensions of threads blocks and the grid
	dim3 dim_block(X_BLOCK, Y_BLOCK, Z_BLOCK);
	dim3 dim_grid(X_GRID, Y_GRID);

	//Launch kernel function with timer
	cudaEvent_t start, stop;
	cudaEventCreate(&start);
	cudaEventCreate(&stop);
	cudaEventRecord(start, 0); 

#ifdef BENCHMARK_POINT_MULTIPLICATION
	ec_multiplication_shared<N><<<dim_grid,dim_block>>>(gpu_data);
#endif
#ifdef EC_DOUBLING
	ec_doubling_shared<N,NB_OPERATIONS><<<dim_grid,dim_block>>>(gpu_data);
#endif
#ifdef EC_ADDITION
	ec_addition_shared<N,NB_OPERATIONS><<<dim_grid,dim_block>>>(gpu_data);
#endif
#ifdef FP_ADDITION
	fp_addition_shared<N,NB_OPERATIONS><<<dim_grid,dim_block>>>(gpu_data);
#endif
#ifdef FP_MULTIPLICATION
	fp_multiplication_shared<N,NB_OPERATIONS><<<dim_grid,dim_block>>>(gpu_data);
#endif


	cudaEventRecord(stop, 0);
	cudaEventSynchronize(stop);
	float elapsedTime;
	cudaEventElapsedTime(&elapsedTime, start, stop);
	cudaEventDestroy(start);
	cudaEventDestroy(stop);

	//copy result from device to host
	get_GPU_data(gpu_data, v_Qx, v_Qy, v_Qz);

	return elapsedTime;
}

/************************************************************************************************/

inline void launch_benchmark(char* benchmark_file, char* result_file) {

	//thread vector of elliptic curve parameters (a,prime, -prime^(-1) mod 2^16)
	vector<Integer<N> > v_prime(NUMBER_THREADS);
	vector<Integer<N> > v_a(NUMBER_THREADS);
	vector<uint>        v_invp(NUMBER_THREADS);

	//thread vector of point P
	vector<Integer<N> > v_Px(NUMBER_THREADS);
	vector<Integer<N> > v_Py(NUMBER_THREADS);
	vector<Integer<N> > v_Pz(NUMBER_THREADS);

	//thread vector of integers with k
	vector<Integer<N> > v_k(NUMBER_THREADS); 

	//thread vector of point [k]P (result)
	vector<Integer<N> > v_Qx(NUMBER_THREADS);
	vector<Integer<N> > v_Qy(NUMBER_THREADS);
	vector<Integer<N> > v_Qz(NUMBER_THREADS);

	//read data from benchmark file
	string param;
	param = read_data(benchmark_file, v_prime, v_a, v_invp, v_Px, v_Py, v_Pz, v_k, v_Qx, v_Qy, v_Qz);

	//run execution 
	float elapsedTime;
	BENCH(1,elapsedTime = run_execution(v_prime, v_a, v_invp, v_Px, v_Py, v_Pz, v_k, v_Qx, v_Qy, v_Qz););

	//write result to the file given as parameter
	write_result(result_file, param, elapsedTime, tps);
}

/************************************************************************************************/

int main (int argc, char* argv[]) {
	cout << "Computation for : " << argv[1] << endl;	
	launch_benchmark(argv[1], argv[2]);
	return 0;
}
